# -*- coding: utf-8 -*-

"""Tessellate module."""

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)
from . import __version__

def createnodejson_forcrd(id, conformer, puckercoords, nextconformer, ringsize, numeric):
    roundedpuckercoords = []
    try:
        for x_ind in puckercoords[1:-2].split(','):
            roundedpuckercoords.append("{0:.2f}".format(float(x_ind)))
    except Exception as e:
        logger.error('%s %s', e, puckercoords)
    jsonobject = {
        'pdbid': id,  # note this could be any id in this case
        'conformer': conformer,
        'puckercoords': roundedpuckercoords,
        'nextconformer': nextconformer,
        'ringsize': ringsize,
        'numeric': numeric
    }
    return jsonobject, "%s %s %s %s %s\n" % (conformer, puckercoords, nextconformer, ringsize, numeric)

def createnodejson(pdbid, resname, chain, resid, ringorder, conformer, contextconformer, puckercoords, nextconformer, ringsize, numeric,macro=False):
    try:
        roundedpuckercoords=[]
        for x_ind in puckercoords:
            roundedpuckercoords.append("{0:.2f}".format(float(x_ind)))
    except Exception as e:
        logger.error('%s %s', e, puckercoords)

    jsonobject = {
        'pdbid': pdbid,
        'resname': resname,
        'chain': chain,
        'resid': resid,
        'ringorder': ringorder,
        'conformer': conformer,
        'contextconformer': contextconformer,
        'puckercoords': roundedpuckercoords,
        'nextconformer': nextconformer,
        'ringsize': ringsize,
        'numeric': numeric,
        'macro' : macro
    }
    return jsonobject, "%s %s %s %s %s %s %s %s %s %s %s\n" % (pdbid, resname, chain, resid, ringorder, conformer, contextconformer, puckercoords, nextconformer, ringsize, numeric)

def analyse_pucker(inputfile, output_dir="", outputfile="tessellate_report", output_format="json"):
    pass
    try:
        import os
        import sys
        import time
        import tessellate.utils.pucker as puc
        import json
        import tessellate.utils.helperfunctions as helperfunctions
    except Exception as e:
        print("Error - Cannot import module %s", e)
        exit(1)

    outputfile = ".".join([outputfile, output_format])
    txt = False
    if output_format == "txt":
        txt = True
        outputfile = open(os.path.join(output_dir, outputfile), 'w')
        outputfile.write("tessellate "+__version__+" txt\n")
        outputfile.write(
            "PUCKER_COORDS CONFORMER ORIG_CONFORMER RING_SIZE\n")
        logger.critical("to be improved, does not match json output at present")
    nodejson = []

    # - does the list of coordinates exist
    inputstream = open(inputfile, 'r')
    for line in inputstream:
        logger.debug('pucker line item %s', line)
        chunked = line.split()
        id = chunked[0]
        ringsize = int((len(chunked) - 1) / 3)
        logger.debug('Ringsize %s', ringsize)
        if ringsize < 5 or ringsize > 8:  # not calculating for these sizes at the moment
            logger.error('Systems containing less than 5 or greater than 8 atoms are not supported. skipping')
            exit(1)  # FIXME!, rather ignore and calculate all others..
        allcoors1 = chunked[1:len(chunked)]
        allcoors = [float(i) for i in allcoors1]
        pobj = puc.Pucker(tuple(allcoors))

        if pobj and pobj.isvalid:
            try:
                thisframeTD = pobj.calculate_triangular_tessellation()
                logger.debug(thisframeTD)
                conformer = pobj.deduce_canonical_conformation()
                nextconformer = pobj.deduce_canonical_conformation(nextguess=True)
                logger.debug('Coordinates are %s %s %s %s : ', str(thisframeTD), str(conformer[0]),
                             str(nextconformer[0]), str(pobj.ringsize))
                node, log = createnodejson_forcrd(id, conformer[0], str(thisframeTD), nextconformer[0], pobj.ringsize, pobj.ttnum[conformer[0]])
                logger.info(log)
                if txt:
                    outputfile.write(log)
                nodejson.append(node)

            except Exception as e:
                logger.error('Pucker object valid, but calc, classify etc. failed  %s %s', str(pobj._coords), e)
                raise e
                pass

    if output_format == "json":
        d2 = [key for key in nodejson]
        helperfunctions.write_to_json(d2, os.path.join(output_dir, outputfile))
    elif output_format == "bson":
        d2 = [key for key in nodejson]
        helperfunctions.write_to_bson(d2, os.path.join(output_dir, outputfile))



def analyse_pucker_from_pdbs(pdbinputfilename, ligandinputfilename=None, outputfile = "tessellate_report",output_format="json", output_dir=""):
    try:
        import Bio.PDB as bp
        import tessellate.utils.pucker as puc
        import numpy as np
        import json
        import tessellate.utils.helperfunctions as helperfunctions
        import tessellate.utils.getRing as getRing
        import os
    except Exception as e:
        print("Error - Cannot import module %s", e)
        exit(1)

    #. workaround stdout annoyance, biopdb sometimes uses print
    try:
        import sys
        from io import StringIO
    except Exception as e:
        print("Error - Cannot import module ", e)
        exit(1)

    all_pucker_json = helperfunctions.init_all_pucker_dictionary()
    all_macro_pucker_json = helperfunctions.init_all_pucker_dictionary()
    #PDBDATA = {}

    outputfile = ".".join([outputfile, output_format])
    txt = False
    if output_format == "txt":
        txt = True
        outputfile = open(os.path.join(output_dir, outputfile), 'w')
        outputfile.write("tessellate "+__version__+" txt\n")
        outputfile.write(
        "PDBID RESNAME CHAIN RESID RINGATOMSORDER CONFORMER CONTEXTUAL_CONFORMER ANGULAR_PUCKER_COORDS ORIG_CONFORMER RING_SIZE\n")
        logger.critical("to be improved, does not match json output at present")
    nodejson = []

    # . read in list of pdbs to read (format is one column of pdb ids
    pdblist = []
    inputfile = open(pdbinputfilename, 'r')
    for line in inputfile:
        logger.debug('Read from pdbnames %s', line)
        pdblist.append(line.strip())

    # . read in ligands with preferred ring ordering (format is name, ringsize, numrings , numrings sets of N atom names
    if ligandinputfilename is None:
        ligand_dict={'3DR': {'num': 1, 'ringids': [("C2'", "C3'", "C4'", "O4'", "C1'")], 'ringsize': 5}, 'AVU': {'num': 2, 'ringids': [("C2'", "C3'", "C4'", "O4'", "C1'"), ('C2R', 'C3R', 'C4R', 'O4R', 'C1R')], 'ringsize': 5}, 'NAG': {'num': 1, 'ringids': [('C3', 'C4', 'C5', 'O5', 'C1', 'C2')], 'ringsize': 6}, 'CTR': {'num': 3, 'ringids': [('C3A', 'C4A', 'C5A', 'O5A', 'C1A', 'C2A'), ('C3B', 'C4B', 'C5B', 'O5B', 'C1B', 'C2B'), ('C3C', 'C4C', 'C5C', 'O5C', 'C1C', 'C2C')], 'ringsize': 6}, 'PSG': {'num': 1, 'ringids': [('C3', 'C4', 'C5', 'O5', 'C1', 'C2')], 'ringsize': 6}, 'MAN': {'num': 1, 'ringids': [('C3', 'C4', 'C5', 'O5', 'C1', 'C2')], 'ringsize': 6}, 'BMA': {'num': 1, 'ringids': [('C3', 'C4', 'C5', 'O5', 'C1', 'C2')], 'ringsize': 6}, 'NDG': {'num': 1, 'ringids': [('C3', 'C4', 'C5', 'O', 'C1', 'C2')], 'ringsize': 6}, 'BGP': {'num': 1, 'ringids': [('C3', 'C4', 'C5', 'O5', 'C1', 'C2')], 'ringsize': 6}, 'G6P': {'num': 1, 'ringids': [('C3', 'C4', 'C5', 'O5', 'C1', 'C2')], 'ringsize': 6}, 'ACR': {'num': 4, 'ringids': [('C3A', 'C4A', 'C5A', 'C7A', 'C1A', 'C2A'), ('C3B', 'C4B', 'C5B', 'O5B', 'C1B', 'C2B'), ('C3C', 'C4C', 'C5C', 'O5C', 'C1C', 'C2C'), ('C3D', 'C4D', 'C5D', 'O5D', 'C1D', 'C2D')], 'ringsize': 6}, 'LAK': {'num': 2, 'ringids': [('C3', 'C4', 'C5', 'O5', 'C1', 'C2'), ("C3'", "C4'", "C5'", "O5'", "C1'", "C2'")], 'ringsize': 6}, 'GAL': {'num': 1, 'ringids': [('C3', 'C4', 'C5', 'O5', 'C1', 'C2')], 'ringsize': 6}, 'BGC': {'num': 1, 'ringids': [('C3', 'C4', 'C5', 'O5', 'C1', 'C2')], 'ringsize': 6}, 'NGA': {'num': 1, 'ringids': [('C3', 'C4', 'C5', 'O5', 'C1', 'C2')], 'ringsize': 6}, 'UPG': {'num': 1, 'ringids': [("C3'", "C4'", "C5'", "O5'", "C1'", "C2'")], 'ringsize': 6}, 'BBA': {'num': 1, 'ringids': [('C1', 'C2', 'C3', 'C', 'C4', 'C5', 'C6')], 'ringsize': 7}, 'H52': {'num': 1, 'ringids': [('N21', 'C22', 'C23', 'N24', 'C25', 'C26', 'C27')], 'ringsize': 7}, '0J0': {'num': 1, 'ringids': [('C15', 'C16', 'C17', 'C18', 'C19', 'C20', 'C21')], 'ringsize': 7}, '13U': {'num': 1, 'ringids': [('C42', 'C43', 'C44', 'C45', 'C46', 'C47', 'C48', 'C49')], 'ringsize': 8}, 'PS9': {'num': 1, 'ringids': [('S2', 'S3', 'S4', 'S5', 'S6', 'S7', 'S8', 'S9')], 'ringsize': 8}}
    else:
        ligfile = open(ligandinputfilename, 'r')
        ligand_dict = {}
        for line in ligfile:
            chunked = line.split()
            id = chunked[0]
            ringsize = int(chunked[1])
            numrings = int(chunked[2])
            columns = int(3)  # columns in addition to atoms
            if len(chunked) == (columns + int(numrings) * int(ringsize)):
                ligand_dict[id] = {}
                ligand_dict[id]["num"] = numrings
                ligand_dict[id]["ringids"] = []
                ligand_dict[id]["ringsize"] = ringsize
                for i in range(0, numrings):
                    templist = []
                    for j in range(0, ringsize):
                        templist.append(chunked[columns + i * ringsize + int(j)])
                    ligand_dict[id]["ringids"].append(tuple(templist))
                logger.debug('Creation of liganddict %s', ligand_dict[id])
            else:
                logger.error('atoms ids relative to number rings does not match in ligands file %s %s %s %s', chunked,
                          len(chunked), numrings, ringsize)
    logger.debug("Ligand Dict used: %s", ligand_dict)

    #. ignore these residues
    aminoacids = ["GLY", "ALA", "SER", "MET", "LYS", "GLU", "PRO", "ASP", "VAL", "PHE", "ASN", "ILE", "TRP",
                          "CYS", "HIS", "LEU", "GLN", "ARG", "TYR", "THR"]
    water = ["HOH"]
    other = []
    soup = aminoacids + water + other
    

    # . download pdbs  and if already there will not download
    for pdbid in pdblist:
        logger.debug('Ids in downloads list %s', pdbid)
        pdbl = bp.PDBList()
        pdbl.retrieve_pdb_file(pdbid, file_format="pdb" , pdir=os.path.join(output_dir,"pdb"))

    # . parse pdbs for ligands, when found, calc pucker
    p = bp.PDBParser()

    # .. loop over all pdbs
    for pdbid in pdblist:
        pdbpath = os.path.join(output_dir,"pdb")+"/pdb" + pdbid.lower() + ".ent"
        # else:
        #     pdbpath = os.path.join(upload_folder,pdbid)
        logger.debug('path to pdbids downloaded %s', pdbid)
        structure = p.get_structure(pdbid, pdbpath)
        # .. get all the residues
        res_list = bp.Selection.unfold_entities(structure, 'R')
        # .. loop over the residues
        for resi in res_list:
            SSSR = []
            rname = resi.get_resname()
            # .. is this residue in the ligands list, calc pucker
            if rname in ligand_dict.keys():
                logger.debug('Ligand %s appears in the ligand dict ', rname)
                # .. get atoms coords
                for nrings in range(0, ligand_dict[rname]["num"]):
                    try:
                        listallcoors = []
                        # .. some pdbs have missing atoms, for this case , check if atoms exist in PDB
                        check_atoms = []
                        for i in resi.get_list():
                            check_atoms.append(i.get_name())
                        missingatomtest = False
                        for atomindex in range(0, ligand_dict[rname]["ringsize"]):
                            if ligand_dict[rname]["ringids"][nrings][atomindex] not in check_atoms:
                                missingatomtest = True
                        if missingatomtest:
                            logger.debug('Missing atoms for %s', rname)
                            pass  # without notifying user !
                        else:
                            for atomindex in range(0, ligand_dict[rname]["ringsize"]):
                                listallcoors = listallcoors + list(
                                    resi[ligand_dict[rname]["ringids"][nrings][atomindex]].get_coord())
                            pobj = puc.Pucker(tuple(listallcoors))
                            if pobj and pobj.isvalid:
                                try:
                                    thisframeTD = pobj.calculate_triangular_tessellation()
                                    logger.debug(thisframeTD)
                                    conformer = pobj.deduce_canonical_conformation()
                                    nextconformer = pobj.deduce_canonical_conformation(nextguess=True)
                                    pconf = pobj.contextualise_conformer(conformer[0],ligand_dict[rname]["ringids"][nrings])
                                    node, log = createnodejson(pdbid, rname, resi.get_parent().get_id(), resi.get_id()[1], ligand_dict[rname]["ringids"][nrings],conformer[0], pconf, thisframeTD, nextconformer[0], pobj.ringsize, pobj.ttnum[conformer[0]])
                                    logger.info(log)
                                    if txt:
                                        outputfile.write(log)
                                    if not nodejson == [] and node in nodejson:
                                        logger.debug('ENTRY EXISTS : %s %s', rname, resi.get_id()[1])
                                    else:
                                        nodejson.append(node)

                                except Exception as e:
                                    logger.error('In known ligs, Pucker object valid, but calc, classify etc. failed  %s %s', str(pobj._coords), e)
                                    raise e
                            else:
                                logger.error("pobj is None or not valid in ligand ring find %s ", listallcoors)
                    except Exception as e:
                        logger.error("pdb file may be missing coordinates..")
                        raise  # not happy about how I raise this error but OK for now

            # .. Calculate SSSR regardless of whether I have this ligand in the dict or not and then calc pucker
            # .. this may make sense but will cause duplicates ......
            if rname not in soup and len(resi.get_list()) > 5:
                atomlist = []
                if resi.is_disordered():
                    for atom in resi.get_list():  # use get_list instead of unpacked list
                        if atom.is_disordered():
                            list_of_disorder = atom.disordered_get_id_list()
                            # just using the last one for now (whatever that is)
                            selected = list_of_disorder[-1]
                            if atom.disordered_has_id(selected):
                                atom.disordered_select(selected)

                for i in resi.get_list():
                    atomlist.append([i.get_name(), i.get_coord()])
                logger.debug("resi %s atomlist %s", resi, atomlist)
                SSSR = getRing.create_graph_and_find_rings_suite(atomlist,mineuclid=1.0,maxeuclid=2.2)

                if SSSR:
                    for ring in SSSR:
                        nring = SSSR.index(ring)
                        logger.debug('FOUND %i rings in resi %s ', nring+1, rname)
                        alpharing = getRing.getcommonring(ring)
                        if alpharing is not None:
                            ring = alpharing
                        else:  # try see if its common through the common dict
                            for common in getRing.commonrings:
                                if sorted(common) == sorted(ring):
                                    ring = common
                                    break
                        listallcoors = []
                        for atomindex in ring:
                            listallcoors = listallcoors + list(resi[atomindex].get_coord())
                        logger.debug("listallcoors %s", listallcoors)
                        pobj = puc.Pucker(tuple(listallcoors))
                        if pobj and pobj.isvalid:
                            try:
                                thisframeTD = pobj.calculate_triangular_tessellation()
                                conformer = pobj.deduce_canonical_conformation()
                                pconf = pobj.contextualise_conformer(conformer[0], ring)
                                nextconformer = pobj.deduce_canonical_conformation(nextguess=True)
                                node, log = createnodejson(pdbid, rname, resi.get_parent().get_id(), resi.get_id()[1], ring, conformer[0], pconf, thisframeTD, nextconformer[0], pobj.ringsize, pobj.ttnum[conformer[0]])
                                logger.info(log)
                                if txt:
                                    outputfile.write(log)
                                if not nodejson == [] and node in nodejson:
                                    logger.debug('ENTRY EXISTS : %s %s', rname, resi.get_id()[1])
                                else:
                                    nodejson.append(node)
                            except Exception as e:
                                logger.error('In SSSR Pucker object valid, but calc, classify etc. failed  %s %s', str(pobj._coords), e)
                                raise e
                        else:
                            logger.error("pobj is None or not valid in SSSR ring find %s %s %s %s", listallcoors,rname,resi.get_id()[1], ring)
                else:
                    logger.debug('%s has no rings', rname)

            #. get all rings in this resi, get com
            macroatomlist = []
            if SSSR:
                for aring in SSSR:
                    ringcoords=[]
                    for ringatom in aring:
                        for itm in resi.get_list():
                            if itm.get_name() == ringatom:
                                ringcoords.append(itm.get_coord())
                    #print aring, np.array(np.add.reduce(ringcoords)/len(ringcoords)), "\n"
                    if len(ringcoords)<9 and len(ringcoords)>4: # ignore too large or too small cycles
                        #print len(ringcoords)
                        macroatomlist.append(["".join(aring), np.array(np.add.reduce(ringcoords)/len(ringcoords))])
                        #macroatomlist.append(["".join(aring), np.array(np.add.reduce(ringcoords))])
                        logger.debug("resi %s rings %s macroatomlist %s", resi, aring, macroatomlist)
                if len(macroatomlist) > 4:  # need at least five cycles to calculate macropucker
                    #print "macroatom ", macroatomlist
                    #import itertools
                    #for a, b in itertools.combinations(macroatomlist, 2):
                    # work out euclidean distance and choose to call this an edge if mineuclid<dist<maxeuclid
                        #print a, b, np.linalg.norm(a[1] - b[1])
                    logger.debug("Macroatoms list %s",macroatomlist)
                    macroSSSR = getRing.create_graph_and_find_rings_suite(macroatomlist,maxeuclid=8.0)
                    logger.debug(macroSSSR)
                    if macroSSSR:
                        #print "mS ", macroSSSR
                        for ring in macroSSSR:
                            nring = macroSSSR.index(ring)
                            logger.debug('FOUND %i macro rings in resi %s ', nring+1, rname)
                            alpharing = getRing.getcommonring(ring)
                            if alpharing is not None:
                                ring = alpharing
                            else:  # try see if its common through the common dict
                                for common in getRing.commonrings:
                                    if sorted(common) == sorted(ring):
                                        ring = common
                                        break
                            listallcoors = []
                            #print macroatomlist
                            for atoms in ring:
                                listallcoors.extend(list((list(x[1]) for x in macroatomlist if x[0] in atoms)))
                            # now flatten the list
                            listallcoors = [y for x in listallcoors for y in x]

                            logger.debug("macro listallcoors %s", listallcoors)
                            pobj = puc.Pucker(tuple(listallcoors))
                            if pobj and pobj.isvalid:
                                try:
                                    thisframeTD = pobj.calculate_triangular_tessellation()
                                    conformer = pobj.deduce_canonical_conformation()
                                    pconf = pobj.contextualise_conformer(conformer[0], ring)
                                    nextconformer = pobj.deduce_canonical_conformation(nextguess=True)
                                    logger.debug("Macrocycles %s %s %s %s", thisframeTD, conformer, pconf, nextconformer)
                                    node, log = createnodejson(pdbid, rname, resi.get_parent().get_id(), resi.get_id()[1], ring, conformer[0], pconf, thisframeTD, nextconformer[0], pobj.ringsize, pobj.ttnum[conformer[0]], True)
                                    logger.info(log)
                                    if txt:
                                        outputfile.write(log)
                                    if not nodejson == [] and node in nodejson:
                                        logger.debug('ENTRY EXISTS : %s %s', rname, resi.get_id()[1])
                                    else:
                                        nodejson.append(node)
                                except Exception as e:
                                    logger.error('In macrocyc, Pucker object valid, but calc, classify etc. failed  %s %s', str(pobj._coords), e)
                                    pass
                            else:
                                logger.error("pobj is None or not valid in macrocyc %s",listallcoors)
                    else:
                        logger.debug('%s has no macro rings', rname)



    if output_format == "json":
        d2 = [key for key in nodejson]
        helperfunctions.write_to_json(d2, os.path.join(output_dir, outputfile))

    inputfile.close()

    if txt:
        outputfile.close()
    return



import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

#| common class ordering of ring systems for 5,6 rings
common5rings = [("C2'", "C3'", "C4'", "O4'", "C1'"), ("C2R", "C3R", "C4R", "O4R", "C1R")]
common6rings = [("C3'", "C4'", "C5'", "O5'", "C1'", "C2'"), ("C3A", "C4A", "C5A", "O5A", "C1A", "C2A"),
                ("C3", "C4", "C5", "O5", "C1", "C2")]
commonrings = common5rings + common6rings


def getcommonring(ring):
    """
    Common atom naming systems for rings, this routine is used to automagically reorder guesses for
    rings s.t. that they provide the expected results...
    """
    thiscycle = {}
    if len(ring) == 5:
        import re
        for atom in ring:  # note this is for a five ring...
            if (atom.startswith('C1') or atom.endswith('C1')) and ('1' in re.findall(r'\d+',atom)): # make sure not matching C13
                thiscycle['C1'] = atom
            elif (atom.startswith('C2') or atom.endswith('C2')) and ('2' in re.findall(r'\d+',atom)): # make sure not matching C23
                thiscycle['C2'] = atom
            elif (atom.startswith('C3') or atom.endswith('C3')) and ('3' in re.findall(r'\d+',atom)): # make sure not matching C33
                thiscycle['C3'] = atom
            elif (atom.startswith('C4') or atom.endswith('C4')) and ('4' in re.findall(r'\d+',atom)): # make sure not matching C44
                thiscycle['C4'] = atom
            elif atom.startswith('O') or atom.endswith('O'):
                thiscycle['O4'] = atom
            else:
                # there is an atom that is not common. So cannot apply this ordering trick.
                return None
        try:
            return (thiscycle['C2'], thiscycle['C3'], thiscycle['C4'], thiscycle['O4'], thiscycle['C1'])
        except:
            return None
    elif len(ring) == 6:
        import re
        for atom in ring:  # note this is for a six ring...
            if (atom.startswith('C1') or atom.endswith('C1')) and ('1' in re.findall(r'\d+',atom)): # make sure not matching C13
                thiscycle['C1'] = atom
            elif (atom.startswith('C2') or atom.endswith('C2')) and ('2' in re.findall(r'\d+',atom)):
                thiscycle['C2'] = atom
            elif (atom.startswith('C3') or atom.endswith('C3')) and ('3' in re.findall(r'\d+',atom)):
                thiscycle['C3'] = atom
            elif (atom.startswith('C4') or atom.endswith('C4')) and ('4' in re.findall(r'\d+',atom)):
                thiscycle['C4'] = atom
            elif (atom.startswith('C5') or atom.endswith('C5')) and ('5' in re.findall(r'\d+',atom)):
                thiscycle['C5'] = atom
            elif atom.startswith('O') or atom.endswith('O'):
                thiscycle['O5'] = atom
            else:
                # what about C10 O9 C14 C13 C12 C11 - should reverse to C11 C12 C13 C14 O9 C10
                # if 10>9 and 10<11 and 11 < 12 then reverse
#                import copy
#                allatoms = " ".join(ring)
#                intsinatoms=re.findall(r'\d+',allatoms)
#                logger.debug("ORDERING %s %s", ring, intsinatoms)
#                if int(intsinatoms[0])> int(intsinatoms[1]) and int(intsinatoms[0])<int(intsinatoms[5]) and int(intsinatoms[5])< int(intsinatoms[4]):
#                    localcopyofring=copy.deepcopy(ring)
#                    #popped=localcopyofring.pop()
#                    #localcopyofring.insert(0,popped)
#                    localcopyofring.reverse()
#                    logger.debug("REVERSE %s", localcopyofring)
#                    return localcopyofring

                #. get O into position 3 (0,1,2,3)
                #. Get O to position 3, then check numbering 4<0
                #!. this depends on atom numbering. If C1 is numbered higher than C4 then the non-contextualised conformer will seem incorrect
                indices = []
                for i, elem in enumerate(ring):
                    if 'O' in elem:
                        indices.append(i)
                if len(indices) == 1:
                    opos=indices[0]
                    if opos==3:
                        return ring
                    pos=opos
                    while pos!=3:
                       popped=ring.pop()
                       ring.insert(0,popped)
                       indices=[]
                       for i, elem in enumerate(ring):
                           if 'O' in elem:
                               indices.append(i)
                       pos=indices[0]
                    logger.debug("POPPEDTOPOS4 %s", ring)
                    try:
                        if int(re.findall(r'\d+',ring[4])[0]) < int(re.findall(r'\d+',ring[0])[0]) :
                            logger.debug("RETURN4 %s", ring)
                            return ring
                        else:
                            popped=ring.pop()
                            ring.insert(0,popped)
                            logger.debug("POPPEDAGAIN %s", ring)
                            return ring
                    except:
                        return None
                else:
                    return None
        try:
            return (
                thiscycle['C3'], thiscycle['C4'], thiscycle['C5'], thiscycle['O5'], thiscycle['C1'], thiscycle['C2'])
        except:
            return None
    else:
        return None


def getring(startatom, atomset):
    """getRing(startatom, atomset, lookup, oatoms)->atoms, bonds
    starting at startatom do a bfs traversal through the atoms
    in atomset and return the smallest ring found

    returns (), () on failure
    note: atoms and bonds are not returned in traversal order"""

    path = {}
    bpaths = {}
    for atomID in atomset.keys():
        # initially the paths are empty
        path[atomID] = []
        bpaths[atomID] = []
    #... insert from Figueras paper
    # The BFS algorithm. We wish to find the smallest ring in the molecule that includes startatom
    # we assign paths , path[i] to each node i. The path conrinas the nodes in the path from the starting node to node i
    # initialise all the paths to null
    # assign values to paths in the starting node
    #for subnodes in atomset[startatom]:
    #   path[subnodes]=[startatom,subnodes]
    # now check if subnodes subnodes is empty

    # but actually will start at starting node but do all nodes...
    q = []
    # Initialize the queue with nodes attached to rootNode
    # and initialise these paths

    for subnodes in atomset[startatom]:
        q.append([startatom, subnodes])
        path[subnodes] = [startatom, subnodes]
    logger.debug("getRING q nodes %s",q)
    logger.debug("getRING path nodes %s",path)
    # loop while the queue size is greater than zero (it exists)
    while q:
        root, node = q.pop()
        for subnodes in atomset[node]:
            logger.debug("getRING subnodes %s in set %s, root is %s", subnodes, atomset[node], root)
            if subnodes != root:  # node shouldn't be start atom but check...
                # check if path is empty or not
                if not path[subnodes]:  # if empty assign path as root path + subnodes
                    path[subnodes] = path[node] + [subnodes]
                    logger.debug("getRING had no paths now path %s from node %s paths %s and itself %s", path[subnodes], node, path[node],[subnodes])
                    q.append([node, subnodes])
                else:  # possible ring closure
                    # compute the intersection of path[root], path[subnodes], it must be a singleton i.e. one element
                    intersection = set(path[node]) & set(path[subnodes])
                    logger.debug("getRING intersection %s", intersection)
                    #print "INTERSECTION ", set(path[node])&set(path[subnodes]), len(intersection)
                    if len(intersection) == int(1):
                        logger.debug("getRING use intersection %s is size %i", intersection, len(intersection))
                        #union
                        union = set(path[node]) | set(path[subnodes])
                        logger.debug("getRING union %s", union)

                        avail = sorted(list(union))  # sort here to prevent dupl
                        logger.debug("getRING avail union %s", avail)
                        chosen = []
                        if len(union) < 5 or len(union)> 400: # it is pointless considering
                            return ()
                        while avail:
                            if not chosen:
                                chosen.append(avail[0])
                                del avail[0]
                            else:
                                lastadded = chosen[-1]
                                for children in atomset[lastadded]:
                                    if children not in chosen and children in avail:  # child must not be used and must be part of the atoms in the ring
                                        chosen.append(children)
                                        del avail[avail.index(children)]
                                        break
                        return chosen
                    else:  # ignore path
                        logger.error("getRING pass on intersection %s is size %i", intersection, len(intersection))
                        pass
            else:
                logger.debug("subnode=root %s %s", subnodes, root)
    logger.debug("returning nothing")
    return ()  # for subnodes in atomSet[startAtom]:


def min_degree(edges):
    """getRing(edges)
    loop through all edges and calculate the min degree of the current nodes in the graph
    returns an int min_degree
    This is this literally the lowest available degree. A graph with 1,2 and 3 degree nodes has a mindegree of 1"""
    min_atom, min_degree = None, int(10000)
    # loop over edges
    for atom in edges:
        if len(edges[atom]) < min_degree:
            min_atom = atom
            min_degree = len(edges[atom])
    return min_atom, min_degree

def create_graph_and_find_rings_suite(atomlist, mineuclid=1.1, maxeuclid=2.0):
    """ find possible rings """
    try:
#        import getRing
        import tessellate.utils.getRing as getRing
    except Exception as e:
       print("Error - Cannot import module ", e)
       exit(1)
    SSSR=[]
    SSSR1=getRing.create_graph_and_find_rings_d3069(atomlist,mineuclid,maxeuclid) 
    SSSR2=getRing.create_graph_and_find_rings_6abb(atomlist,mineuclid,maxeuclid)
    SSSR.extend(SSSR1)
    for itm in SSSR2:
        if itm not in SSSR:
            SSSR.extend([itm])
    return SSSR

def make_unique(original_list):
    unique_list = []
    map(lambda x: unique_list.append(x) if (x not in unique_list) else False, original_list)
    return unique_list

def create_graph_and_find_rings_d3069(atomlist, mineuclid=1.1, maxeuclid=2.0):
    """
    :type atomlist: list
    :type mineuclid: float
    :type maxeuclid: float
    :rtype : list
    """
    try:
        #import getRing
        import tessellate.utils.getRing as getRing
        import itertools
        import numpy as np
    except Exception as e:
        print("Error - Cannot import module ", e)
        exit(1)
    SSSR = [] # keep track of all the rings
    edges = {}
    for a, b in itertools.combinations(atomlist, 2):
        # work out euclidean distance and choose to call this an edge if mineuclid<dist<maxeuclid
        dist = np.linalg.norm(a[1] - b[1])
        if maxeuclid > dist > mineuclid:
            try:
                edges[a[0]].append(b[0])
            except:
                edges[a[0]] = [b[0]]
            try:
                edges[b[0]].append(a[0])
            except:
                edges[b[0]] = [a[0]]

    # debug print out the edges
    for atom in edges:
        logger.debug('Atom edges info %s %s %s', atom, edges[atom], len(edges[atom]))
    # Now recursively remove all terminal nodes i.e.with only one edge
    alledges = dict(edges)
    deletededges = {}
    #logger.debug("allededges dict %s",alledges)
    #logger.debug("edges %s",edges)
    if edges != []:
        while edges:
            for atom in dict(edges).keys():
                N2nodes = []
                N3nodes = []
                logger.debug("all edge keys %s",edges.keys())
                logger.debug("Current edge %s",atom)
                if atom in deletededges.keys():
                    logger.critical("revisiting a deleted edge")
                    exit(1)
                if int(len(edges[atom])) == int(0):  # trim degree zero nodes
                    logger.debug("Zero edge node deleted %s",atom)
                    deletededges[atom]=edges[atom]
                    del edges[atom]  # cannot delete dictionary while iterating use a while
                elif int(len(edges[atom])) == int(1):  # trim degree one nodes
                    # pop it from other atoms
                    logger.debug("One edge node %s with %i edges %s removed from parent node edges list %s",atom, len(edges[atom]),edges[atom],edges[edges[atom][0]])
                    edges[edges[atom][0]].remove(atom)
                    edges[atom] = []
                    deletededges[atom]=edges[atom]
                    del edges[atom]  # cannot delete dictionary while iterating use a while
                elif int(len(edges[atom])) == int(2):  # find nodes of degree 2 and add to N2 nodes
                    logger.debug("Two edge node %s with %i edges",atom, len(edges[atom]))
                    N2nodes.append(atom)
                elif int(len(edges[atom])) == int(3):  # find nodes of degree 2 and add to N3 nodes
                    logger.debug("Three edge node %s with %i edges",atom, len(edges[atom]))
                    N3nodes.append(atom)

                if getRing.min_degree(edges)[1] == int(2) and N2nodes:  # the minimum degree of the entire graph
                    logger.debug("current min degree %i", getRing.min_degree(edges)[1] )
                    for N2 in N2nodes:
                        ring = getRing.getring(N2, alledges)  # give this ring the entire graph
                        logger.debug("ring from getring %s", ring)
                        if len(ring) > 0:
                            if ring in SSSR:  # if exists then ignore, not sorting as a unique ring is being offered by getRing
                                logger.error("ring already in SSSR %s %i %s", ring, len(ring),type(ring))
                                pass
                            else:
                                SSSR.append(ring)
                        else:
                            pass
                        try: # try isolate and eliminate one N2 node
                            logger.debug("isolate N2 node %s with %i edges and break one bond ", N2, len(edges[N2]) )
                            bondtobreak=edges[N2].pop()
                            logger.debug("break %s with current edges %s ", bondtobreak, edges[bondtobreak] )
                            edges[bondtobreak].remove(N2)
                        except:
                            logger.error("N2 not eliminated")
                            pass

                elif getRing.min_degree(edges)[1] == int(3) and N3nodes:  # the minimum degree of the entire graph
                    logger.debug("current min degree %i", getRing.min_degree(edges)[1] )
                    logger.debug("N3 d3069")
                    for N3 in N3nodes:
                        ring = getRing.getring(N3, alledges)  # give this ring the entire graph
                        logger.debug("ring from getring %s", ring)
                        if len(ring) > 0:
                            if ring in SSSR:  # if exists then ignore, not sorting as a unique ring is being offered by getRing
                                logger.debug("ring already in SSSR %s %i %s", ring, len(ring),type(ring))
                                pass
                            else:
                                SSSR.append(ring)
                        else:
                            pass
                        try: # try isolate and eliminate one N3 node
                            logger.debug("isolate N3 node %s with %i edges and break one bond ", N3, len(edges[N3]) )
                            bondtobreak=edges[N3].pop()
                            logger.debug("break %s with current edges %s ", bondtobreak, edges[bondtobreak] )
                            edges[bondtobreak].remove(N3)
                        except Exception as e:
                            logger.error(e)
                            print(e)
    if SSSR:
        logger.debug("SSSR %s", SSSR)
    return SSSR



def create_graph_and_find_rings_old(atomlist, mineuclid=1.1, maxeuclid=2.0):
    """
    :type atomlist: list
    :type mineuclid: float
    :type maxeuclid: float
    :rtype : list
    """
    try:
        # import getRing
        import tessellate.utils.getRing as getRing
        import itertools
        import numpy as np
    except Exception as e:
        print("Error - Cannot import module ", e)
        exit(1)
    SSSR = [] # keep track of all the rings
    edges = {}
    for a, b in itertools.combinations(atomlist, 2):
        # work out euclidean distance and choose to call this an edge if mineuclid<dist<maxeuclid
        dist = np.linalg.norm(a[1] - b[1])
        if maxeuclid > dist > mineuclid:
            try:
                edges[a[0]].append(b[0])
            except:
                edges[a[0]] = [b[0]]
            try:
                edges[b[0]].append(a[0])
            except:
                edges[b[0]] = [a[0]]

    for atom in edges:
        logger.debug('Atom edges info %s %s %s', atom, edges[atom], len(edges[atom]))
    # Now recursively remove all terminal nodes i.e.with only one edge
    alledges = dict(edges)
    if edges != []:
        while edges:
            for atom in dict(edges).keys():
                N2nodes = []
                logger.debug("all edge keys %s",edges.keys())
                logger.debug("Current edge %s",atom)
                if int(len(edges[atom])) == int(0):  # trim degree zero nodes
                    logger.debug("Zero edge node deleted %s",atom)
                    del edges[atom]  # cannot delete dictionary while iterating use a while
                elif int(len(edges[atom])) == int(1):  # trim degree one nodes
                    # pop it from other atoms
                    logger.debug("One edge node %s with %i edges %s removed from parent node edges list %s",atom, len(edges[atom]),edges[atom],edges[edges[atom][0]])
                    edges[edges[atom][0]].remove(atom)
                    edges[atom] = []
                    del edges[atom]  # cannot delete dictionary while iterating use a while
                elif int(len(edges[atom])) == int(2):  # find nodes of degree 2 and add to N2 nodes
                    logger.debug("Two edge node %s with %i edges",atom, len(edges[atom]))
                    N2nodes.append(atom)
                if getRing.min_degree(edges)[1] == int(2):  # the minimum degree of the entire graph
                    logger.debug("current min degree %i", getRing.min_degree(edges)[1] )
                    for N2 in N2nodes:
                        ring = getRing.getring(N2, alledges)  # give this ring the entire graph
                        logger.debug("ring from getring %s", ring)
                        if len(ring) > 0:
                            if ring in SSSR:  # if exists then ignore, not sorting as a unique ring is being offered by getRing
                                logger.debug("ring already in SSSR %s %i %s", ring, len(ring),type(ring))
                                pass
                            else:
                                SSSR.append(ring)
                        else:
                            pass
                        try: # try isolate and eliminate one N2 node
                            logger.debug("isolate N2 node %s with %i edges and break one bond ", N2, len(edges[N2]) )
                            bondtobreak=edges[N2].pop()
                            logger.debug("break %s with current edges %s ", bondtobreak, edges[bondtobreak] )
                            edges[bondtobreak].remove(N2)
                        except:
                            logger.error("Couldn't eliminate N2")
                            pass

                elif getRing.min_degree(edges)[1] == int(3):  # the minimum degree of the entire graph
                    logger.debug("current min degree %i", getRing.min_degree(edges)[1] )
                    ring = getRing.getring(atom, alledges)
                    logger.debug("ring from getring %s", ring)
                    if len(ring) > 0:
                        if ring in SSSR:
                            pass
                        else:
                            SSSR.append(ring)
                    else:
                        pass
                    try: # select an optimum edge for elimination. trial each edge in alledges
                        logger.error("FUTURETODO - N3 check edges not yet implemented only applicable to cages etc." )
                        exit(1)
                    except:
                        pass
    if SSSR:
        logger.debug("SSSR %s", SSSR)
    return SSSR

def create_graph_and_find_rings_6abb(atomlist, mineuclid=1.1, maxeuclid=2.0):
    """
    :type atomlist: list
    :type mineuclid: float
    :type maxeuclid: float
    :rtype : list
    """
    try:
        # import getRing
        import tessellate.utils.getRing as getRing
        import itertools
        import numpy as np
    except Exception as e:
        print("Error - Cannot import module ", e)
        exit(1)
    SSSR = [] # keep track of all the rings
    edges = {}
    for a, b in itertools.combinations(atomlist, 2):
        # work out euclidean distance and choose to call this an edge if mineuclid<dist<maxeuclid
        dist = np.linalg.norm(a[1] - b[1])
        if maxeuclid > dist > mineuclid:
            #edges.append([a,b]) # this works but is difficult to remove edges later
            try:
                edges[a[0]].append(b[0])
            except:
                edges[a[0]] = [b[0]]
            try:
                edges[b[0]].append(a[0])
            except:
                edges[b[0]] = [a[0]]

    for atom in edges:
        logger.debug('Atom edges info %s %s %s', atom, edges[atom], len(edges[atom]))
    # Now recursively remove all terminal nodes i.e.with only one edge
    alledges = dict(edges)
    deletededges = {}
    if edges != []:
        while edges:
            N2nodes = []
            N3nodes = []
            for atom in dict(edges).keys():
                logger.debug("all edge keys %s",edges.keys())
                logger.debug("Current edge %s",atom)
                if atom in deletededges.keys():
                    logger.critical("revisiting a deleted edge")
                    exit(1)

                if int(len(edges[atom])) == int(0):  # trim degree zero nodes
                    logger.debug("Zero edge node deleted %s",atom)
                    deletededges[atom]=edges[atom]
                    del edges[atom]  # cannot delete dictionary while iterating use a while
                elif int(len(edges[atom])) == int(1):  # trim degree one nodes
                    # pop it from other atoms
                    logger.debug("One edge node %s with %i edges %s removed from parent node edges list %s",atom, len(edges[atom]),edges[atom],edges[edges[atom][0]])
                    edges[edges[atom][0]].remove(atom)
                    edges[atom] = []
                    deletededges[atom]=edges[atom]
                    logger.critical("Deleting atom %s %s",atom, deletededges)
                    del edges[atom]  # cannot delete dictionary while iterating use a while
                elif int(len(edges[atom])) == int(2):  # find nodes of degree 2 and add to N2 nodes
                    logger.debug("Two edge node %s with %i edges",atom, len(edges[atom]))
                    N2nodes.append(atom)
                elif int(len(edges[atom])) == int(3):  # find nodes of degree 2 and add to N3 nodes
                    logger.debug("Three edge node %s with %i edges",atom, len(edges[atom]))
                    N3nodes.append(atom)

            if getRing.min_degree(edges)[1] == int(2) and N2nodes:  # the minimum degree of the entire graph
                logger.debug("current min degree %i", getRing.min_degree(edges)[1] )
                for N2 in N2nodes:
                    ring = getRing.getring(N2, alledges)  # give this ring the entire graph
                    logger.debug("ring from getring %s", ring)
                    if len(ring) > 0:
                        if ring in SSSR:  # if exists then ignore, not sorting as a unique ring is being offered by getRing
                            logger.debug("ring already in SSSR %s %i %s", ring, len(ring),type(ring))
                        else:
                            SSSR.append(ring)
                logger.debug("isolate N2 node %s with %i edges and break one bond ", N2nodes[0], len(edges[N2nodes[0]]) )
                bondtobreak=edges[N2nodes[0]].pop(0)
                logger.debug("break %s with current edges %s ", bondtobreak, edges[bondtobreak] )
                edges[bondtobreak].remove(N2nodes[0])

            elif getRing.min_degree(edges)[1] == int(3) and N3nodes:  # the minimum degree of the entire graph
                logger.debug("current min degree %i", getRing.min_degree(edges)[1] )
                logger.debug("N3 6abb")
                for N3 in N3nodes:
                    ring = getRing.getring(N3, alledges)  # give this ring the entire graph
                    logger.debug("ring from getring %s", ring)
                    if len(ring) > 0:
                        if ring in SSSR:  # if exists then ignore, not sorting as a unique ring is being offered by getRing
                            logger.debug("ring already in SSSR %s %i %s", ring, len(ring),type(ring))
                            pass
                        else:
                            SSSR.append(ring)
                    else:
                        pass
                    try: # try isolate and eliminate one N3 node
                        logger.debug("isolate N3 node %s with %i edges and break one bond ", N3, len(edges[N3]) )
                        bondtobreak=edges[N3].pop()
                        logger.debug("break %s with current edges %s ", bondtobreak, edges[bondtobreak] )
                        edges[bondtobreak].remove(N3)
                    except Exception as e:
                        logger.error(e)
                        print(e)
    if SSSR:
        logger.debug("SSSR %s", SSSR)
    return SSSR
